#
# deploy.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *
import nemo

def _get_eps_at(self, *args, **kwargs):
    if hasattr(self, 'graph'):
        if self.graph is not None:
            return self.graph.get_eps_at(*args, **kwargs)

def _harden_weights_pact(self, **kwargs):
    r"""Harden all weights in the network to their quantized value.

    """

    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            m.train_loop_oldprec = float(m.W_beta.item()+m.W_alpha.item())/(2.0**(m.W_precision.get_bits())-1)
            m.harden_weights(**kwargs)
        if (m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            m.harden_weights(**kwargs)

def _round_weights_pact(self, **kwargs):
    r"""Round all weights in the network adding 1/2 an eps.

    """

    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            m.weight.data[:] += (m.W_beta.item()+m.W_alpha.item())/(2.0**(m.W_precision.get_bits())-1) / 2
        if (m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            m.kappa.data[:] += m.eps_kappa/2
            m.lamda.data[:] += m.eps_lamda/2

def _set_deployment_pact(self, eps_in, only_activations=False, **kwargs):
    r"""Sets the network in deployment mode, enabling saving it to ONNX format or similar.

    :param eps_in: Input precision quantum.
    :type  eps_in: float

    """

    self.stage = 'qd'
    if not only_activations:
        self.eps_in = eps_in
        self.set_eps_in(eps_in)
    for n,m in self.named_modules():
        if ((not only_activations and m.__class__.__name__ == "PACT_Conv2d") or \
            (not only_activations and m.__class__.__name__ == "PACT_Conv1d") or \
            (not only_activations and m.__class__.__name__ == "PACT_Linear") or \
            (not only_activations and m.__class__.__name__ == "PACT_IntegerAdd") or \
                                      m.__class__.__name__ == "PACT_Act"):
            m.deployment = True
        if (m.__class__.__name__ == "PACT_Act"):
            m.set_static_precision(**kwargs)

def _set_eps_in_pact(self, eps_in):
    r"""Sets the input precision quantum of the network.

    :param eps_in: Input precision quantum.
    :type  eps_in: float

    """

    assert(hasattr(self, 'graph'))
    assert(self.graph is not None)
    self.graph.rebuild_module_dict()
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Act"):
            m.eps_in = torch.tensor(self.get_eps_at(n, eps_in).item(), requires_grad=False)
        if (m.__class__.__name__ == "PACT_QuantizedBatchNormNd"):
            m.eps_in = torch.tensor(self.get_eps_at(n, eps_in).item(), requires_grad=False)
        if (m.__class__.__name__ == "PACT_IntegerAdd"):
            eps_in_tmp = self.get_eps_at(n, eps_in)
            eps_in_list = []
            for eps in eps_in_tmp:
                eps_in_list.append(torch.tensor(eps.item(), requires_grad=False))
            m.eps_in_list = eps_in_list

def _qd_stage(self, eps_in, add_input_bias_dict={}, remove_bias_dict={}, prune_empty_bn=True, int_accurate=True, bn_calibration_fn=None, bn_calibration_range_factor=8, **kwargs):
    r"""High-level function to move the network from FQ to QD stage.

    :param eps_in: Input precision quantum (required).
    :type  eps_in: float
    :param add_input_bias_dict: dictionary of layers to which an input bias must be added (layer name as key, bias as value).
    :type  add_input_bias_dict: dict or `collections.OrderedDict`
    :param remove_bias_dict: dictionary of Linear->BatchNorm couples where bias must be absorbed by the BatchNorm (Linear name as key, BatchNorm name as value).
    :type  remove_bias_dict: dict or `collections.OrderedDict`
    :param prune_empty_bn: if True (default), BatchNorm channel multiplicative parameters are pruned if smaller than 1e-9.
    :type  prune_empty_bn: bool
    :param int_accurate: if True (default), target an accurate representation of ID numerical dynamics (e.g., with requantization) at QD stage.
    :type  int_accurate: bool
    :param bn_calibration_fn: if not None (default), a function (e.g., calling validation) used to calibrate BatchNorm range.
    :type  bn_calibration_fn: function
    :param bn_calibration_range_factor: if bn_calibration_fn is None, multiply the clipping parameter of the following Activation multiplied by bn_calibration_range_factor to estimate BatchNorm range.
    :type  bn_calibration_range_factor: int

    """

    if prune_empty_bn:
        self.prune_empty_bn(threshold=1e-9)
    self.round_weights()
    self.harden_weights()
    if add_input_bias_dict:
        self.add_input_bias(add_input_bias_dict)
    if remove_bias_dict:
        self.remove_bias(remove_bias_dict)
    if int_accurate:
        nemo.transform.bn_quantizer(self, **kwargs)
    else:  # this is mainly useful for debug purposes, to identify misalignments FQ/QD
        for n,m in self.named_modules():
            if (m.__class__.__name__ == "PACT_Act"):
                m.precise = True
    self.set_deployment(eps_in=eps_in, **kwargs) # with initial BN eps
    if bn_calibration_fn is not None:
        with self.statistics_bn():
            bn_calibration_fn()
        self.calibrate_bn(**kwargs)
    else:
        self.calibrate_bn(minmax=False, range_factor=bn_calibration_range_factor, **kwargs)
    self.set_deployment(eps_in=eps_in, **kwargs) # repeat, to fix BN eps
    self.harden_weights()

def _id_stage(self, eps_in=None, **kwargs):
    r"""High-level function to move the network from QD to ID stage.

    :param eps_in: Input precision quantum, default None (will use the previously saved eps_in).
    :type  eps_in: float

    """

    if self.stage == 'fq':
        self.qd_stage(eps_in=eps_in, **kwargs)
    self.stage = 'id'
    if eps_in is None:
        eps_in = self.eps_in
    nemo.transform.integerize_pact(self, eps_in=eps_in, **kwargs)




#
# relaxation.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
# Alfio Di Mauro <adimauro@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import logging
import nemo
import torch
from collections import OrderedDict
from tqdm import tqdm

class RelaxationEngine():
    r"""Main engine class for weight/activation precision relaxation procedure.

        :param net: Module or network on which the relaxation procedure will be performed (must have a `change_precision` method).
        :type  net: `torch.nn.Module`

        :param optimizer: A pointer to the optimizer being used for the training.
        :type  optimizer: `torch.optim.Optimizer`

        :param criterion: Loss function being used as a quality metric.
        :type  criterion: `torch.autograd.Function`

        :param trainloader: Loader used for training values (used for relaxation evaluation).
        :type  trainloader: `torch.utils.data.DataLoader`

        :param precision_rule: A dictionary describing the rules to be used for the precision relaxation.
        :type  precision_rule: `dict` or `collections.OrderedDict`

        :param tbx_writer: TensorBoardX writer for data logging. Defaults to `None`.
        :type  tbx_writer: `tensorboardx.SummaryWriter`

        :param reset_alpha_weights: If True, reset the W_alpha and W_beta parameters at precision change.
        :type  reset_alpha_weights: `bool`

        :param min_prec_dict: Dictionary of minimum allowed precision for all parameters.
        :type  min_prec_dict: `dict` or `collections.OrderedDict`

        :param evaluator: Evaluation engine for precision selection heuristics.
        :type  evaluator: `nemo.evaluation.EvaluationEngine`

        :param evaluator_threshold: Threshold to be used for precision binning (default 0.9).
        :type  evaluator_threshold: `float`

        :param evaluator_verbose: If True, print more information from evaluation engine.
        :type  evaluator_verbose: `bool`

        :param evaluator_strategy: Can be 'max_accuracy' (default) or 'min_precision'.
        :type  evaluator_strategy: `string`

        :param divergence_policy: Can be 'change_precision' (default) or 'lower_lr'.
        :type  divergence_policy: `string`

        :param divergence_lr_scaler: LR scaling factor for 'lower_lr' divergence policy.
        :type  divergence_lr_scaler: `float`

        The :py:class:`RelaxationEngine` represents a precision-scaling procedure utilizing the relaxation
        heuristic.
        The main parameters of the relaxation heuristic are passed through the `precision_rule`
        dictionary when initializing the class.

        Assuming `p` is the `precision_rule`, the relaxation heuristic keeps an internal memory
        of the loss and its relative variation (`delta_loss`) over the last `p[running_avg_memory]` epochs.
        The mean and standard deviation of `delta_loss` over `p[running_avg_memory]` epochs are compared with 
        `p[delta_loss_less_than]` and `p[delta_loss_running_std_stale]`. Moreover the absolute loss value
        is compared with `p[abs_loss_stale]`.
        
        Two counters are updated:
        - for each consecutive epoch in which all the three values compared are less than the respective
        parameters from `p`, an `abs_bound` counter is updated
        - for each consecutive epoch in which the two delta values compared are less than the respective
        parameters from `p`, a `no_abs_bound` counter is updated

        After the counters are updated, their respective value is compared with `p[for_epochs]` and
        `p[for_epochs_no_abs_bound]`: if any of the counters is higher than the respective parameter, the
        training is considered "stale" for the current quantization value.

        When this happens, precision is scaled down by a factor of `p[bit_scaler]` bits, up to the point
        when `p[bit_stop_condition]` precision is reached. If `p[scale_lr]` is set to true, the learning
        rate is also downscaled by a factor of `p[lr_scaler]`.
    """

    def __init__(self, net, optimizer, criterion, trainloader, precision_rule=None, tbx_writer=None, reset_alpha_weights=True, min_prec_dict=None, evaluator=None, evaluator_threshold=0.9, evaluator_verbose=False, evaluator_strategy='min_precision', divergence_policy='change_precision', divergence_lr_scaler=0.2, evaluate=None, log_start=None, log_stop=None, log_step=None, validate_on_train_fn=None, reset_alpha_below_6bits=False):
        super(RelaxationEngine, self).__init__()

        self.precision_rule = precision_rule
        if self.precision_rule is None:
            self.scale_activations = True
            self.scale_weights = True
        else:
            try:
                self.scale_activations = self.precision_rule['scale_activations']
            except KeyError:
                self.scale_activations = True
            try:
                self.scale_weights = self.precision_rule['scale_weights']
            except KeyError:
                self.scale_weights = True

        if self.precision_rule is not None:
            self.delta_loss_running_avg = 1000.0
            self.delta_loss_running_std = 0.0
            self.delta_loss_memory_curr_size = 0
            self.delta_loss_memory_size = self.precision_rule['running_avg_memory']
            self.delta_loss_memory = np.zeros(self.precision_rule['running_avg_memory'])
            
            try:
                cs = self.precision_rule['custom_scaler']
            except KeyError:
                self.precision_rule['custom_scaler'] = None

        self.loss_best = 1e3
        self.loss_prev = 1e3
        self.net = net
        self.optimizer = optimizer
        self.precision_abs_bound_counter = 0
        self.precision_no_abs_bound_counter = 0
        self.criterion = criterion
        self.trainloader = trainloader

        self.tbx_writer = tbx_writer
        self.reset_alpha_weights = reset_alpha_weights

        self.min_prec_dict = min_prec_dict
        self.evaluator = evaluator
        self.evaluator_threshold = evaluator_threshold
        self.evaluator_verbose = evaluator_verbose
        self.evaluator_strategy = evaluator_strategy

        self.divergence_policy = divergence_policy
        self.divergence_lr_scaler = divergence_lr_scaler
        self.divergence_cnt = 0
        self.divergence_lrscaling_limit = 3

        self.validate_on_train_fn = validate_on_train_fn

        self.reset_alpha_below_6bits = reset_alpha_below_6bits

        self.relaxation_ended = False

    def step(self, loss, epoch=0, checkpoint_name='checkpoint', previous_loss=None):
        r"""Iterate a step over the relaxation engine, checking the current convergence rate and updating precision and LR.

            :param loss: Current value of the training loss function.
            :type  loss: `torch.Tensor`

            :param epoch: Epoch of training.
            :type  epoch: `int`

            :param checkpoint_name: String to be used as a name for the checkpoint file.
            :type  checkpoint_name: `str`

            The `step` method iterates the `RelaxationEngine` using as input the current value of the training loss function,
            whose convergence is evaluated.
        """

        reset_alpha_weights = self.reset_alpha_weights
        if self.precision_rule is None:
            return

        # save best result in case of catastrophic failure
        if loss < self.loss_best:
            self.loss_best = loss
            if checkpoint_name is not None:
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name=checkpoint_name, checkpoint_suffix="_current_best")
            else:
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name="checkpoint", checkpoint_suffix="_current_best")

        try:
            curr_regime = self.precision_rule[epoch]
        except KeyError:
            try:
               curr_regime = self.precision_rule[str(epoch)]
            except KeyError:
               curr_regime = None

        for i in range(self.delta_loss_memory_size-1, 0, -1):
            self.delta_loss_memory[i] = self.delta_loss_memory[i-1]
        self.delta_loss_memory[0] = self.loss_prev-loss
        if self.delta_loss_memory_curr_size < self.delta_loss_memory_size:
            self.delta_loss_memory_curr_size += 1
        if self.delta_loss_memory_curr_size > 1:
            delta_loss_running_avg = self.delta_loss_memory[:self.delta_loss_memory_curr_size-1].mean()
            delta_loss_running_std = self.delta_loss_memory[:self.delta_loss_memory_curr_size-1].std()
        else:
            delta_loss_running_avg = np.Inf
            delta_loss_running_std = np.Inf
        logging.info("[Relax @%d]\t delta_loss_running_avg=%.3e loss_epoch_m1=%.3e delta_loss=%.3e" % (epoch-1, delta_loss_running_avg, loss, self.loss_prev-loss))
        logging.info("[Relax @%d]\t delta_loss_memory=%s" % (epoch-1, self.delta_loss_memory[:self.delta_loss_memory_curr_size-1]))

        if self.tbx_writer is not None:
            self.tbx_writer.add_scalars('train', { 'delta_loss_avg': delta_loss_running_avg, 'delta_loss_std': delta_loss_running_std }, epoch+1)

        # staleness happens when 1) delta_loss has bounded mean and std and absolute loss is bounded for for_epochs, or 2) delta_loss has bounded mean and std for for_epochs_no_abs_bound
        if delta_loss_running_avg < self.precision_rule['delta_loss_less_than'] and delta_loss_running_std < self.precision_rule['delta_loss_running_std_stale'] and loss < self.precision_rule['abs_loss_stale']:
            self.precision_abs_bound_counter += 1
        else:
            self.precision_abs_bound_counter = 0
        if delta_loss_running_avg < self.precision_rule['delta_loss_less_than'] and delta_loss_running_std < self.precision_rule['delta_loss_running_std_stale']:
            self.precision_no_abs_bound_counter += 1
        else:
            self.precision_no_abs_bound_counter = 0

        divergence_chprec_flag = False
        divergence_lowlr_flag = False
        # catastrophic failure occurs when delta_loss_running_avg is negative
        try:
            divergence_abs_threshold = self.precision_rule['divergence_abs_threshold']
        except KeyError:
            divergence_abs_threshold = 1e9
        if delta_loss_running_avg < 0 or loss > self.precision_rule['divergence_abs_threshold']:
            # recover previous state
            state = torch.load("checkpoint/"+checkpoint_name+"_current_best.pth")['state_dict']
            self.net.load_state_dict(state, strict=True)
            logging.info("[Relax @%d]\t Detected divergent training, restoring previous best state." % (epoch-1))

            if (self.divergence_policy == 'lower_lr' and self.divergence_cnt < self.divergence_lrscaling_limit) or self.relaxation_ended:
                # reset delta loss memory (with a small delta)
                self.precision_abs_bound_counter = 0
                self.precision_no_abs_bound_counter = 0
                loss = self.loss_best
                self.delta_loss_memory[:] = 0
                self.delta_loss_memory_curr_size = 1
                self.divergence_cnt += 1
                if self.divergence_cnt == 1:
                    self.divergence_saved_lr = list(self.optimizer.param_groups)[0]['lr']

                # scale LR together with W bits
                for p in self.optimizer.param_groups:
                    p['lr'] *= self.divergence_lr_scaler
                    lr = p['lr']
                logging.info("[Relax @%d]\t Using 'lower_lr' policy (iter %d); scaled LR to %.3e" % (epoch-1, self.divergence_cnt, lr))

                # report this as a precision change
                divergence_lowlr_flag = True
            elif self.divergence_policy == 'change_precision' or self.divergence_cnt >= self.divergence_lrscaling_limit:
                divergence_chprec_flag = True
                self.divergence_cnt = 0
                logging.info("[Relax @%d]\t Using 'change_precision' policy." % (epoch-1))
            else:
                raise NotImplementedError

        # if relaxation has already ended, exit here
        if self.relaxation_ended:
            return divergence_lowlr_flag, True

        change_precision = divergence_lowlr_flag # False in normal cases

        if curr_regime is not None:
            self.precision_abs_bound_counter = 0
            self.precision_no_abs_bound_counter = 0
            if loss is None:
                loss = 1e3
            self.loss_prev = 1e3
            suffix  = '_' + "%.1f" % (self.net.W_precision.get_bits()) + 'b'
            suffix += 'x' + "%.1f" % (self.net.x_precision.get_bits()) + 'b'
            if checkpoint_name is not None:
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name=checkpoint_name, checkpoint_suffix=suffix)
            self.net.change_precision(bits=curr_regime['W_bits'], scale_activations=False, scale_weights=True, reset_alpha=reset_alpha_weights, min_prec_dict=self.min_prec_dict)
            self.net.change_precision(bits=curr_regime['x_bits'], scale_activations=True, scale_weights=False, reset_alpha=reset_alpha_weights, min_prec_dict=self.min_prec_dict)

            # save checkpoint for catastrophic failure case
            self.loss_best = 1e3
            if checkpoint_name is None:
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name='checkpoint', checkpoint_suffix="_current_best")
            else:
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name=checkpoint_name, checkpoint_suffix="_current_best")

        elif self.precision_abs_bound_counter == self.precision_rule['for_epochs'] or self.precision_no_abs_bound_counter == self.precision_rule['for_epochs_no_abs_bound'] or divergence_chprec_flag:

            if self.precision_abs_bound_counter == self.precision_rule['for_epochs']:
                logging.info("[Relax @%d]\t precision_abs_bound_counter=%d: Triggering precision change below absolute loss threshold" % (epoch-1, self.precision_abs_bound_counter))
            elif self.precision_no_abs_bound_counter == self.precision_rule['for_epochs_no_abs_bound']:
                logging.info("[Relax @%d]\t precision_no_abs_bound_counter=%d: Triggering precision change above absolute loss threshold" % (epoch-1, self.precision_no_abs_bound_counter))

            self.precision_abs_bound_counter = 0
            self.precision_no_abs_bound_counter = 0
            loss = 1e3
            self.loss_prev = 2e3
            self.delta_loss_memory[:] = 1e3
            self.delta_loss_memory_curr_size = 1

            try:
                scale_x = self.precision_rule['scale_x']
            except KeyError:
                scale_x = True
            try:
                scale_W = self.precision_rule['scale_W']
            except KeyError:
                scale_W = True

            # stop condition is currently measured against W_precision.bits
            if (self.net.W_precision.get_bits() >= self.precision_rule['W_bit_stop_condition']) or \
               (self.net.x_precision.get_bits() >= self.precision_rule['x_bit_stop_condition']):

                # save checkpoint
                if checkpoint_name is not None:
                    suffix  = '_' + "%.1f" % (self.net.W_precision.get_bits()) + 'b'
                    suffix += 'x' + "%.1f" % (self.net.x_precision.get_bits()) + 'b'
                    nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name=checkpoint_name, checkpoint_suffix=suffix)

                # if there is an EvaluationEngine available, use it
                if self.evaluator is not None:
                    W_start = self.net.W_precision.get_bits()
                    x_start = self.net.x_precision.get_bits()
                    # pure heuristics...
                    if W_start > 8:
                        W_step = -2
                        W_stop = max(6, self.precision_rule['W_bit_stop_condition']-2) if scale_W else W_start+W_step
                    elif W_start > 6:                      
                        W_step = -1
                        W_stop = max(5, self.precision_rule['W_bit_stop_condition']-1) if scale_W else W_start+W_step
                    else:
                        W_step = -0.5
                        W_stop = max(3.5, self.precision_rule['W_bit_stop_condition']-0.5) if scale_W else W_start+W_step
                    if x_start > 8:
                        x_step = -2
                        x_stop = max(6, self.precision_rule['x_bit_stop_condition']-2) if scale_x else x_start+x_step
                    elif x_start > 6:
                        x_step = -1
                        x_stop = max(5, self.precision_rule['x_bit_stop_condition']-1) if scale_x else x_start+x_step
                    else:
                        x_step = -0.5
                        x_stop = max(3.5, self.precision_rule['x_bit_stop_condition']-0.5) if scale_x else x_start+x_step
                    self.net.unset_train_loop() # this is a "soft" weight hardening
                    self.evaluator.reset_grids(W_start, W_stop, W_step, x_start, x_stop, x_step)
                    while self.evaluator.step():
                        acc = self.evaluator.validate_fn(0, val_loader=self.evaluator.validate_data)
                        self.evaluator.report(acc)
                        logging.info("[Relax @%d]\t %.1f-bit W, %.1f-bit x %.2f%%" % (epoch-1, self.evaluator.wgrid[self.evaluator.idx], self.evaluator.xgrid[self.evaluator.idx], 100*acc.item()))
                    self.net.set_train_loop() # this removes the "soft" hardening
                    Wbits, xbits = self.evaluator.get_next_config(upper_threshold=self.evaluator_threshold, verbose=self.evaluator_verbose, strategy=self.evaluator_strategy)
                    Wdiff = W_start - Wbits
                elif self.precision_rule['custom_scaler'] is not None:
                    scaler = self.precision_rule['custom_scaler']
                    if len(scaler) == 0:
                        return True, True
                    Wbits, xbits, lrscaled, divpol = scaler.pop(0)
                    self.divergence_policy = divpol # update divergence policy
                    W_diff = self.net.W_precision.get_bits() - Wbits
                else:
                    Wdiff = -self.precision_rule['W_bit_scaler']
                    Wbits = self.net.W_precision.get_bits()+self.precision_rule['W_bit_scaler']
                    xbits = self.net.x_precision.get_bits()+self.precision_rule['x_bit_scaler']
                    
                logging.info("[Relax @%d]\t Choosing %.1f-bit W, %.1f-bit x for next step" % (epoch-1, Wbits, xbits))
                if scale_W:
                    self.net.change_precision(bits=Wbits, scale=self.net.W_precision.get_scale(), scale_activations=False, scale_weights=self.scale_weights,     reset_alpha=reset_alpha_weights, min_prec_dict=self.min_prec_dict)
                    # this will reset alpha,beta PACT parameters to 5 standard deviations upon precision change, to avoid wasting dynamic range to represent irrealistic weights
                    if Wbits < 6 and self.reset_alpha_below_6bits:
                        self.net.reset_alpha_weights(stdev=5.)
                        logging.info("[Relax @%d]\t Setting alpha,beta params of weights to %.1f std deviations" % (epoch-1, 5))
                if scale_x:
                    self.net.change_precision(bits=xbits, scale=self.net.x_precision.get_scale(), scale_activations=self.scale_activations, scale_weights=False, reset_alpha=reset_alpha_weights, min_prec_dict=self.min_prec_dict)
                change_precision = True
                if self.divergence_cnt > 0:
                    for p in self.optimizer.param_groups:
                        p['lr'] = self.divergence_saved_lr
                self.divergence_cnt = 0
                try:
                    if self.evaluator is not None:
                        # scale LR together with W bits
                        for p in self.optimizer.param_groups:
                            p['lr'] *= (2**Wdiff)
                            lr = p['lr']
                        logging.info("[Relax @%d]\t Scaled LR to %.3e" % (epoch-1, lr))
                    elif self.precision_rule['custom_scaler'] is not None:
                        for p in self.optimizer.param_groups:
                            p['lr'] = lrscaled
                            lr = p['lr']
                        logging.info("[Relax @%d]\t Scaled LR to %.3e" % (epoch-1, lr))
                    elif self.precision_rule['scale_lr']:
                        for p in self.optimizer.param_groups:
                            p['lr'] *= self.precision_rule['lr_scaler']
                            lr = p['lr']
                        logging.info("[Relax @%d]\t Scaled LR to %.3e" % (epoch-1, lr))
                except KeyError:
                    pass
                if self.validate_on_train_fn is not None:
                    loss = self.validate_on_train_fn(epoch)
                    logging.info("[Relax @%d]\t validate_on_train loss=%.3e" % (epoch-1, loss.item()))

                # save checkpoint for catastrophic failure case
                self.loss_best = loss
                nemo.utils.save_checkpoint(self.net, self.optimizer, epoch-1, checkpoint_name=checkpoint_name, checkpoint_suffix="_current_best")

            else:
                self.relaxation_ended = True
                logging.info("[Relax @%d]\t Precision relaxation procedure ended" % (epoch-1))
                return True, True
            
        if self.tbx_writer is not None:
            self.tbx_writer.add_scalars('train', { 'abs_bound_counter': self.precision_abs_bound_counter, 'no_abs_bound_counter': self.precision_no_abs_bound_counter, }, epoch+1)
            lr_save = list(self.optimizer.param_groups)[0]['lr']
            self.tbx_writer.add_scalars('train', { 'lr': lr_save }, epoch+1)
        logging.info("[Relax @%d]\t delta_loss_running_avg=%.3e loss_epoch_m1=%.3e delta_loss=%.3e" % (epoch-1, delta_loss_running_avg, loss, self.loss_prev-loss))
        logging.info("[Relax @%d]\t delta_loss_memory=%s" % (epoch-1, self.delta_loss_memory[:self.delta_loss_memory_curr_size-1]))
        logging.info("[Relax @%d]\t precision_abs_bound_counter=%d precision_no_abs_bound_counter=%d" % (epoch-1, self.precision_abs_bound_counter, self.precision_no_abs_bound_counter))

        self.loss_prev = loss

        # if precision has changed, signal this upstream
        return change_precision, False



#
# common.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re

__all__ = [ "reshape_before", "reshape_after", "weight_range", "weight_min", "weight_max", "onnx_name_2_pytorch_name", "get_bn_dict_from_supernodes", "get_equalize_dict_from_supernodes", "get_calib_dict_from_supernodes" ]

def reshape_before(m, s):
    if   m.__class__.__name__ == "PACT_Conv2d":
        return s.reshape((s.shape[0],1,1,1))
    elif m.__class__.__name__ == "PACT_Conv1d":
        return s.reshape((s.shape[0],1,1))
    elif m.__class__.__name__ == "PACT_Linear":
        return s.reshape((s.shape[0],1))
    else:
        return s

def reshape_after(m, s):
    if   m.__class__.__name__ == "PACT_Conv2d":
        if m.groups == s.shape[0]:
            # dwc
            return s.reshape((s.shape[0],1,1,1))
        else:
            return s.reshape((1,s.shape[0],1,1))
    elif m.__class__.__name__ == "PACT_Conv1d":
        return s.reshape((1,s.shape[0],1))
    elif m.__class__.__name__ == "PACT_Linear":
        return s.reshape((1,s.shape[0]))
    else:
        return s

def weight_max(m, range_idx):
    if   m.__class__.__name__ == "PACT_Conv2d":
        if m.groups == m.weight.shape[0]:
            range_idx = 0 # for DW-conv, always marginalize idx 1
        return m.weight.max(3)[0].max(2)[0].max(1-range_idx)[0] 
    elif m.__class__.__name__ == "PACT_Conv1d":
        return m.weight.max(2)[0].max(1-range_idx)[0]
    elif m.__class__.__name__ == "PACT_Linear":
        return m.weight.max(1-range_idx)[0]
    elif m.__class__.__name__ == "BatchNorm1d" or m.__class__.__name__ == "BatchNorm2d":
        return m.weight.data.abs()

def weight_min(m, range_idx):
    if   m.__class__.__name__ == "PACT_Conv2d":
        if m.groups == m.weight.shape[0]:
            range_idx = 0 # for DW-conv, always marginalize idx 1
        return m.weight.min(3)[0].min(2)[0].min(1-range_idx)[0]
    elif m.__class__.__name__ == "PACT_Conv1d":
        return m.weight.min(2)[0].min(1-range_idx)[0]
    elif m.__class__.__name__ == "PACT_Linear":
        return m.weight.min(1-range_idx)[0]
    elif m.__class__.__name__ == "BatchNorm1d" or m.__class__.__name__ == "BatchNorm2d":
        return m.weight.data.abs()

def weight_range(m, range_idx, symmetric=False):
    if   m.__class__.__name__ == "PACT_Conv2d":
        if m.groups == m.weight.shape[0]:
            range_idx = 0 # for DW-conv, always marginalize idx 1
        if not symmetric:
            return m.weight.max(3)[0].max(2)[0].max(1-range_idx)[0] - m.weight.min(3)[0].min(2)[0].min(1-range_idx)[0]
        else:
            return 2*m.weight.abs().max(3)[0].max(2)[0].max(1-range_idx)[0]
    elif m.__class__.__name__ == "PACT_Conv1d":
        if not symmetric:
            return m.weight.max(2)[0].max(1-range_idx)[0] - m.weight.min(2)[0].min(1-range_idx)[0]
        else:
            return m.weight.abs().max(2)[0].max(1-range_idx)[0]
    elif m.__class__.__name__ == "PACT_Linear":
        if not symmetric:
            return m.weight.max(1-range_idx)[0] - m.weight.min(1-range_idx)[0]
        else:
            return m.weight.abs().max(1-range_idx)[0]
    elif m.__class__.__name__ == "BatchNorm1d" or m.__class__.__name__ == "BatchNorm2d":
        return m.weight.data.abs()

def onnx_name_2_pytorch_name(name):
    name_parts = re.findall('\[.*?\]', name)
    name_parts = [part[1:-1] for part in name_parts]
    return '.'.join(name_parts)
 
def get_equalize_dict_from_supernodes(net):
    eq_dict = OrderedDict()
    act_dict = OrderedDict()
    # check all supernodes for ACT and CONV layers
    lin = {}
    act = {}
    prev = {}
    for k,ssn in net.graph.get_supernodes().items():
        for n in ssn['supernode']:
             if isinstance(n[1], PACT_Conv2d) or \
                isinstance(n[1], PACT_Conv1d) or \
                isinstance(n[1], PACT_Linear):
                 lin[k]  = n[0]
        prev[k] = ssn['previous']
        act[k]  = k
    for k in prev.keys():
        p = lin.get(prev[k])
        if p is not None:
            eq_dict [lin[prev[k]]] = lin[k]
            act_dict[lin[prev[k]]] = act[prev[k]]
    return eq_dict, act_dict

def get_bn_dict_from_supernodes(net):
    bn_dict = OrderedDict()
    # check all supernodes for BN and CONV layers
    for k,ssn in net.graph.get_supernodes().items():
        bn = []
        lin = []
        sn = ssn['supernode']
        prev = ssn['previous']
        for n in sn:
            if isinstance(n[1], torch.nn.BatchNorm2d) or \
               isinstance(n[1], torch.nn.BatchNorm1d) or \
               isinstance(n[1], PACT_QuantizedBatchNormNd):
                bn.append(n[0])
            if isinstance(n[1], PACT_Conv2d) or \
               isinstance(n[1], PACT_Conv1d) or \
               isinstance(n[1], PACT_Linear):
                lin.append(n[0])
        if len(lin) > 1 or len(bn) > 1:
            print("[Error] Supernode analysis identified multiple BN or LIN layers in supernode")
            print(lin, bn)
            return
        try:
            bn_dict[lin[0]] = bn[0]
        except IndexError:
            pass
    return bn_dict

def get_calib_dict_from_supernodes(net):
    calib_dict = OrderedDict()
    # check all supernodes for BN and ACT layers
    for k,ssn in net.graph.get_supernodes().items():
        bn = []
        sn = ssn['supernode']
        prev = ssn['previous']
        for n in sn:
            if isinstance(n[1], torch.nn.BatchNorm2d) or \
               isinstance(n[1], torch.nn.BatchNorm1d) or \
               isinstance(n[1], PACT_QuantizedBatchNormNd):
                bn.append(n[0])
        if len(bn) > 1:
            print("[Error] Supernode analysis identified multiple BN layers in supernode")
            print(bn)
            return
        try:
            calib_dict[bn[0]] = k
        except IndexError:
            pass
    return calib_dict




#
# statistics.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *
from contextlib import contextmanager

@contextmanager
def _statistics_act_pact(self):
    r"""Used with `with net.statistics_act():`, calls `net.set_statistics_act()` on enter
    and `net.unset_statistics_act()` on exit.    

    """
    self.set_statistics_act()
    try:
        yield
    finally:
        self.unset_statistics_act()

@contextmanager
def _statistics_bn_pact(self):
    r"""Used with `with net.statistics_bn():`, calls `net.set_statistics_bn()` on enter
    and `net.unset_statistics_bn()` on exit.    

    """
    self.set_statistics_bn()
    try:
        yield
    finally:
        self.unset_statistics_bn()

def _set_statistics_act_pact(self):
    r"""Sets :py:class:`nemo.quant.PACT_Act` layers to collect statistics and work like ReLU's.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_Act":
            m.statistics_only = True

def _get_statistics_act_pact(self):
    r"""Returns the statistics collected by :py:class:`nemo.quant.PACT_Act` layers.
    
    """

    d = OrderedDict([])
    for n,m in self.named_modules():
        d[n] = OrderedDict([])
        if m.__class__.__name__ == "PACT_Act":
            d[n]['max']          = m.get_statistics()[0]
            d[n]['running_mean'] = m.get_statistics()[1]
            d[n]['running_var']  = m.get_statistics()[2]
            d[n]['active']       = m.statistics_only
    return d

def _unset_statistics_act_pact(self):
    r"""Sets :py:class:`nemo.quant.PACT_Act` layers to act normally and stop statistics collection.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_Act":
            m.statistics_only = False

def _set_statistics_bn_pact(self):
    r"""Sets :py:class:nemo.quant.PACT_QuantizedBatchNormNd` layers to collect statistics.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_QuantizedBatchNormNd":
            m.statistics_only = True

def _unset_statistics_bn_pact(self):
    r"""Sets :py:class:`nemo.quant.PACT_QuantizedBatchNormNd` layers to act normally and stop statistics collection.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_QuantizedBatchNormNd":
            m.statistics_only = False



#
# bias.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *

def _add_input_bias_pact(self, lin_dict, eps_in=None):
    r"""Adds a bias to compensate the asymmetry of an input activation tensor.
    
    :param lin_dict: a dictionary, where the key is the linear layer name and the value is the related activation translation.
    :type  lin_dict: `dict` or `collections.OrderedDict`

    """

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "ConstantPad2d"):
            module_dict[n] = m
    # print(lin_dict)
    for n in lin_dict.keys():
        m = module_dict[n]
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            try:
                m.bias.data[:] = m.bias.data[:] - lin_dict[n] * m.weight.data[:].sum(3).sum(2).sum(1)
            except AttributeError:
                m.bias = torch.nn.Parameter(-lin_dict[n] * m.weight.data[:].sum(3).sum(2).sum(1))
            if eps_in is None:
                m.padding_value = lin_dict[n]
            else:
                m.padding_value = math.floor(lin_dict[n]/eps_in)*eps_in
        elif eps_in is None:
            m.value = lin_dict[n]
        else:
            m.value = math.floor(lin_dict[n]/eps_in)*eps_in
    self.input_bias_dict = lin_dict

def _remove_input_bias_pact(self):
    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            module_dict[n] = m
    for n in self.input_bias_dict.keys():
        m = module_dict[n]
        m.bias.data[:] = m.bias.data[:] + self.input_bias_dict[n] * m.weight.data[:].sum(3).sum(2).sum(1)
        m.padding_value = 0
    self.input_bias_dict = None

def _remove_bias_pact(self, bn_dict={}):
    r"""Folds the bias of a linear layer into the parameters of a following batch-norm.
    
    :param bn_dict: a dictionary of layer names, with the key being the source (linear) and the value the target (batch-norm).
    :type  bn_dict: `dict` or `collections.OrderedDict`

    """

    if not bn_dict:
        bn_dict = get_bn_dict_from_supernodes(self)

    module_dict = {}
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" or \
            m.__class__.__name__ == "BatchNorm2d" or \
            m.__class__.__name__ == "BatchNorm1d" ):
            module_dict[n] = m
    for n_before in bn_dict.keys():
        n_after = bn_dict[n_before]
        m_before = module_dict[n_before]
        m_after  = module_dict[n_after]
        m_after.running_mean.data[:] -= m_before.bias.data[:]
        m_before.bias = None




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytorch-nemo",
    version="0.0.7",
    author="Francesco Conti",
    author_email="f.conti@unibo.it",
    description="NEural Minimizer for pytOrch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pulp-platform/nemo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        "torch>=1.3.0",
        "torchvision>=0.4.1",
        "numpy",
        "tqdm",
        "packaging",
        "scikit-learn"
    ]
)



#
# export.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *

def _export_precision(self):
    r"""Returns a dictionary of precisions for each layer.

    """

    d = OrderedDict([])
    for n,m in self.named_modules():
        d[n] = OrderedDict([])
        try:
            d[n]['x_bits']  = m.precision.get_bits()
            if not "PACT" in m.__class__.__name__:
                d[n]['x_scale'] = m.precision.get_scale()
        except AttributeError:
            pass
        try:
            d[n]['W_bits']  = m.W_precision.get_bits()
            if not "PACT" in m.__class__.__name__:
                d[n]['W_scale'] = m.W_precision.get_scale()
        except AttributeError:
            pass
        if len(d[n].keys()) == 0 or n == "":
            d.pop(n, None)
    return d

def _export_weights_legacy_int16(self, header_name='weights.h', save_binary=False, folder_name='.', x_alpha_safety_factor=1.):
    r"""Exports weights and bias values with the legacy strategies used e.g. in PULP-DroNet,
    towards INT-16. Quantization is fully symmetric and aligned to power-of-two `alpha` so that
    there is no need to propagate :math:`\varepsilon` values.
    
    :param header_name: name of a header file.
    :type  header_name: string
    :param save_binary: if True, saves also a binary version.
    :type  save_binary: bool
    :param folder_name: name of the folder where to save binaries.
    :type  folder_name: string

    """

    weight_dict = {}
    bias_dict = {}
    qi_dict = {} # actually qi-1, as 1 bit is for sign
    x_alpha = 0.001
    W_alpha = {}
    bigstr = "/* weights & biases */\n\n#include <stdint.h>\n\n\n"
    checkstr_w = ""
    checkstr_b = ""
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Act"):
            x_alpha = max(x_alpha, m.alpha.item())
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Act"):
            m.alpha.data[:] = 2.**int(math.ceil(math.log2(x_alpha)))
    x_alpha *= x_alpha_safety_factor
    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            W_alpha[n] = max(-m.W_alpha.item(), m.W_beta.item())
            qi_dict[n] = int(math.ceil(math.log2(W_alpha[n])))
            W_eps = 2.**-(16-qi_dict[n]-1)
            m.W_beta.data[:]  = 2.**qi_dict[n]
            m.W_alpha.data[:] = 2.**qi_dict[n]
            # performs also weight hardening, destructive!
            m.harden_weights()
            weight_dict[n] = np.int16(((m.weight.data.clone().detach().to('cpu').numpy()) / W_eps))
            m.weight.data[:] = torch.tensor(weight_dict[n] * W_eps)
            x_eps = 2.**-(16-int(math.ceil(math.log2(x_alpha)))-1)
            try:
                bias_dict[n]   = np.int16(m.bias.data.clone().detach().to('cpu').numpy() / x_eps)
                m.bias.data[:] = torch.tensor(bias_dict[n] * x_eps)
            except AttributeError:
                bias_dict[n]   = np.int16(np.zeros(weight_dict[n].shape[0]))
            import re
            n_str = re.sub('[^0-9a-zA-Z_]+', '_', n)
            bigstr += "// %s weights [shape=%s, qi=%d, qf=%d]\n" %  (n, weight_dict[n].shape, qi_dict[n]+1, 16-qi_dict[n]-1)
            bigstr += "int16_t w_%s[] = {\n  " % n_str
            for i in range(len(weight_dict[n].flatten())-1):
                bigstr += "0x%04x,\n  " % np.uint16(weight_dict[n].flatten()[i])
            bigstr += "0x%04x\n};\n\n" % np.uint16(weight_dict[n].flatten()[-1])
            bigstr += "// %s bias [shape=%s, qi=%d, qf=%d]\n" %  (n, bias_dict[n].shape, int(math.ceil(math.log2(x_alpha)))+1, 16-int(math.ceil(math.log2(x_alpha)))-1)
            bigstr += "int16_t b_%s[] = {\n  " % n_str
            for i in range(len(bias_dict[n].flatten())-1):
                bigstr += "0x%04x,\n  " % np.uint16(bias_dict[n].flatten()[i])
            bigstr += "0x%04x\n};\n\n\n" % np.uint16(bias_dict[n].flatten()[-1])
            if save_binary:
                with open("%s/weights_%s.hex" % (folder_name, n_str), "w") as file:
                    weight_dict[n].flatten().tofile(file)
                with open("%s/bias_%s.hex" % (folder_name, n_str), "w") as file:
                    bias_dict[n].flatten().tofile(file)
                checkstr_w += "Checksum weights_%s:\t%s\n" % (n_str, sum(weight_dict[n].flatten()))
                checkstr_b += "Checksum bias_%s:\t%s\n" % (n_str, sum(bias_dict[n].flatten()))
    print("Export procedure completed, qi=%d qf=%d for activations" % (int(math.ceil(math.log2(x_alpha)))+1, 16-int(math.ceil(math.log2(x_alpha)))-1))
    with open("%s/%s" % (folder_name, header_name), "w") as file:
        file.write(bigstr)
    with open("%s/checksum.txt" % (folder_name), "w") as file:
        file.write(checkstr_w)
        file.write(checkstr_b)


#
# utils.py
# Francesco Conti <fconti@iis.ee.ethz.ch>
#
# Copyright (C) 2018-2020 ETH Zurich
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import nemo
from nemo.precision import Precision
from nemo.quant.pact import *
from nemo.graph import DeployGraph
from torch.nn.modules.utils import _single,_pair
from collections import OrderedDict
import types
import logging
import numpy as np
import copy
import math
import torchvision.models
import re
from nemo.transf.common import *

def _change_precision_pact(self, bits=4, scale_activations=True, scale_weights=True, verbose=True, reset_alpha=True, min_prec_dict=None, **kwargs):
    r"""Changes the target precision of a PACT quantization-aware layer.

    
    :param bits: target bit-width.
    :type  bits: `int`

    :param scale_activations: if False, do not change precision of activations (default True).
    :type  scale_activations: boolean

    :param scale_weights: if False, do not change precision of weights (default True).
    :type  scale_weights: boolean

    :param verbose: if False, do not log precision information (default True).
    :type  verbose: boolean

    :param reset_alpha: if False, do not reset weight scale parameter upon precision change (default True).
    :type  reset_alpha: boolean

    :param min_prec_dict: dictionary of minimum layer-by-layer precisions (default None).
    :type  min_prec_dict: dictionary

    """
    if scale_activations and bits is not None:
        self.x_precision.bits = bits
    if scale_weights and bits is not None:
        self.W_precision.bits = bits
    for n,m in self.named_modules():
        min_prec_x = copy.deepcopy(self.x_precision)
        min_prec_W = copy.deepcopy(self.W_precision)
        if min_prec_dict is not None:
            try:
                min_prec_x.bits = min_prec_dict[n]['x_bits']
            except KeyError:
                pass
            try:
                min_prec_W.bits = min_prec_dict[n]['W_bits']
            except KeyError:
                pass
        if m.__class__.__name__ == "PACT_Act" and scale_activations:
            m.precision = max(self.x_precision, min_prec_x)
        if scale_weights and (m.__class__.__name__ == "PACT_Conv2d" or \
                              m.__class__.__name__ == "PACT_Conv1d" or \
                              m.__class__.__name__ == "PACT_Linear"):
            m.W_precision = max(self.W_precision, min_prec_W)
            if reset_alpha:
                m.reset_alpha_weights()
        if verbose and (m.__class__.__name__ == "PACT_Act") and scale_activations:
            try:
                logging.info("[Quant]\t\t %s: x_bits=%.2f" % (n, m.precision.get_bits()))
            except AttributeError:
                pass
        if verbose and scale_weights and (m.__class__.__name__ == "PACT_Conv2d" or \
                                          m.__class__.__name__ == "PACT_Conv1d" or \
                                          m.__class__.__name__ == "PACT_Linear"):
            try:
                logging.info("[Quant]\t\t %s: W_bits=%.2f" % (n, m.W_precision.get_bits()))
            except AttributeError:
                pass

def _set_train_loop_pact(self):
    r"""Sets modules so that weights are not treated like hardened (e.g., for training).
    
    """

    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" ):
            m.train_loop = True

def _unset_train_loop_pact(self):
    r"""Sets modules so that weights are treated like hardened (e.g., for evaluation).
    
    """

    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear" ):
            m.train_loop = False
            m.train_loop_oldprec = float(m.W_beta.item()+m.W_alpha.item())/(2.0**(m.W_precision.get_bits())-1)

def _reset_alpha_act_pact(self, **kwargs):
    r"""Resets :py:class:`nemo.quant.PACT_Act` parameter `alpha` the value collected through statistics.
    
    """

    for n,m in self.named_modules():
        if m.__class__.__name__ == "PACT_Act":
            m.reset_alpha(**kwargs)

def _get_nonclip_parameters_pact(self):
    r"""Yields all parameters except for `alpha` values.
    
    """

    for name, param in self.named_parameters(recurse=True):
        if name[-5:] != 'alpha':
            yield param

def _get_clip_parameters_pact(self):
    r"""Yields all `alpha` parameters.
    
    """

    for name, param in self.named_parameters(recurse=True):
        if name[-5:] == 'alpha':
            yield param

def _reset_alpha_weights_pact(self, method='standard', **kwargs):
    r"""Resets parameter `W_alpha`.
    
    """

    for n,m in self.named_modules():
        if (m.__class__.__name__ == "PACT_Conv2d" or \
            m.__class__.__name__ == "PACT_Conv1d" or \
            m.__class__.__name__ == "PACT_Linear"):
            m.reset_alpha_weights(**kwargs)

